
/*
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */

package com.sunlabs.archsim.fleet.ships.stride;

import com.sunlabs.archsim.base.hierarchy.CompositeEntity;
import com.sunlabs.archsim.base.simulation.Command;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.fleet.common.Data;
import com.sunlabs.archsim.fleet.common.OutputFunction;
import com.sunlabs.archsim.gasp.TriggerAction;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;

/**
 * A counting action produces the counts in a stride component.
 * The counting is triggered by execution of the enable trigger commend.
 * Parammeters to the enable trigger command are packed into BeginCountRecord,
 * which contains the start value for the counter, the number of steps to take,
 * and the counter increment for each step.
 * 
 *  The counter value is provided on output counterTerm (which is paired with
 *  acknowledegment input ackCounterTerm).
 *  A counter value is produced only upon receipt of an input on nextTerm (paired with
 *  output ackNextTerm).
 *  When the count is done (all steps completed), the CountingAction executes a doneCommand.
 *  This command must be provided at componnent creation and is used to inform a DoneAction
 *  that the count has completed and a "done" token can be produced.
 * 
 * @author ib27688
 *
 */

class CountingAction extends TriggerAction {

	
	/* input - token to produce the next counter */
	private Terminal nextTerm= null;
	/* output - LOCAL ack that the token has been received */
	private Terminal ackNextTerm= null;
	
	/* output - the counter */
	private Terminal counterTerm= null;
	/** input - LOCAL ack that the counter has been received */
	private Terminal ackCounterTerm;


	/** starting vale for the counter*/
	private int counterStart;
	/** current counter value */
	private int counterValue;
	/** counter step */
	private int counterStep;
	/** is the countter valid */
	private boolean counterValid;
	
	/** remnaining number of count for the stride */
	private int countStepsLeft;

	/** command to be invoked when the current count is done */
	Command doneCommand=null;
	
	/** The function to "compute" the acknowledgment value */
	OutputFunction ack;
	/** The Function to compute the next value of the counter */
	OutputFunction count;
	
	public CountingAction(String name, Command done) {
		super(name);
		
		// make the output functions
		ack= new AckFunction();
		count= new Count();
		// associate the output functions with terminal names
		outputFunctions.put(StrideDefaults.ACK_NEXT, ack);
		outputFunctions.put(StrideDefaults.COUNTER, count);
		
		// set the done command
		doneCommand= done;
	}

	public CountingAction(String name, Command done, CompositeEntity g) {
		// make the action
		this(name, done);
		// join the group
		g.addMember(this);
	}
	
	/** Attach a terminal:
	 * Accept only the terminal names from StrideDefaults:
	 * NEXT, ACK_NEXT, COUNTER, ACK_COUNTER are treated as generic terminals 
	 * 
	 * The follwing terminals are, however, treated specially: 
	 * START_TRIGGER_IN, ACK_START_TRIGGER_IN, 
	 * 
	 * Input on START_TRIGGER_IN delivers the counter, the step, 
	 * and the number of steps. This data resets the counting.
	 * If this input is received before the previous count was
	 * complete, 
	 *  
	 */
	public void attach(Terminal t) {
		// get the bas ename for check
		String termName= t.getBaseName();
		// remember whether the terminal was of a bad type
		boolean badType= false;
		// remember whether the same name been added twice?
		boolean twice= false;

		if (termName.equals(StrideDefaults.NEXT)) { 
			if (nextTerm == null) {
				if (t instanceof InputTerminal) {
					// attach
					t.attachAction(this);
					numInputs++;
					// remember what terminal this is
					nextTerm= (InputTerminal)t;
					// add the terminal to the set
					terminals.add(t);
				} 
				else { 
					badType= true;
				}
			}
			else {
				twice= true;
			}
		}
		else if (termName.equals(StrideDefaults.ACK_NEXT)) { 
			if (ackNextTerm == null) {
				if (t instanceof OutputTerminal) {
					t.attachAction(this);
					ackNextTerm= (OutputTerminal)t;
					terminals.add(t);
				} 
				else { 
					badType= true;
				}
			}
			else {
				twice= true;
			}
		}
		else if (termName.equals(StrideDefaults.COUNTER)) { 
			if (counterTerm == null) {
				if (t instanceof OutputTerminal) {
					t.attachAction(this);
					counterTerm= (OutputTerminal)t;
					terminals.add(t);
				} 
				else { 
					badType= true;
				}
			}
			else {
				twice= true;
			}
		}
		else if (termName.equals(StrideDefaults.ACK_COUNTER)) { 
			if (ackCounterTerm == null) {
				if (t instanceof InputTerminal) {
					t.attachAction(this);
					numInputs++;
					ackCounterTerm= (InputTerminal)t;
					terminals.add(t);
				} 
				else { 
					badType= true;
				}
			}
			else {
				twice= true;
			}
		}
		else {
			// unknown terminal
			director.error("Terminal " 
					+ t.getName() 
					+ " cannot be attached to CountingAction "
					+ getName() + ", id= " + getID()
					+ "; terminal name not recognized");
		}

			
		// check for errors
		if (badType) {
			// the terminal was of a wrong type
			director.fatalError("CountingAction, " + getName() + ", " + getID() +  
					": Terminal " + t.getName() + " is of a bad type");
		}
		if (twice) {
			// this was the second addition of the same name
			director.fatalError("CountingAction, " + getName() + ", " + getID() 
					+ ": Terminal " + t.getName() + " cannot be attached twice");
		}
		
	} // attach
	
	/** 
	 * Enable trigger packs the initial value for the counter,
	 * step size, and the number of steps.
	 * 
	 * @param params
	 */
	protected void processEnableTrigger(Object params) {
		// re-enabling is not permitted, this is an error
		if (enabled) {
			director.fatalError("CountingAction "
					+ getName() + ", id= " + getID()
					+ ", was reenabled with the remaining count of "
					+ countStepsLeft
					);
		}
		BeginCountRecord r= (BeginCountRecord)params;
		counterStart= r.startValue;
		counterValue= counterStart;
		counterStep= r.stepSize;
		countStepsLeft= r.numberOfSteps;
		counterValid= r.valid;
		// for a valid counter, there should be at least one step
		if (countStepsLeft <= 0) counterValid= false;
		// is the start value valid?
		// if not, the counter is not valid
		if (counterStart > StrideDefaults.MAX_COUNTER 
				|| counterStart < StrideDefaults.MIN_COUNTER)
			counterValid= false;
	} // processEnnableTrigger
		
	
	/** 
	 * This fuction computes the next count.
	 * When the count ends, it schdeules the "done" command.
	 */ 
	class Count extends OutputFunction {

		Delay triggerDelay;
		
		public Count() {
			super("Count");
			/** get the function delay - for producing the next count */
			Integer fDelay= globals.intValue(StrideDefaults.STRIDE_COUNTER_DELAY);
			functionDelay= (fDelay != null)
					? new Delay(fDelay.intValue())
					: StrideDefaults.STRIDE_FUNCTION_DELAY_DEF;
			
			/** get the triggert delay - for generating a done trigger */
			fDelay= globals.intValue(StrideDefaults.STRIDE_TRIGGER_DELAY);
			triggerDelay= (fDelay != null)
					? new Delay(fDelay.intValue())
					: StrideDefaults.STRIDE_TRIGGER_DELAY_DEF;
			
		}
	
		// Here we update the counter
		public Data value() {
			Data newValue= null;
			// is the counter valid?
			if (!counterValid) {
				newValue= new Data(false);
				// this is the last step, no matter what
				// note this to disable the action later on
				countStepsLeft= 0;
			}
			else {
				// remember the current counter, for output
				newValue= new Data(counterValue);
				// this step is taken, so decrease the count to see whether this is the last
				// step
				countStepsLeft--;
				
				// if there are more steps left compute the next counter, check for validity
				if (countStepsLeft > 0) {
					if (	 (counterStep>0 && (StrideDefaults.MAX_COUNTER-counterStep) < counterValue)
						|| (counterStep<0 && (StrideDefaults.MIN_COUNTER-counterStep) > counterValue)) {
						// the counter would go out of range in the next iteration
						counterValid= false;
					}
					else {
						// the next value of counter will be valid, so we can compute it
						counterValue += counterStep;
					}
				} // if countStepsLeft > 0	
			} // else (the counter is valid upon entry)

			// if the count is done, produce a notification and disable the action
			if (countStepsLeft <= 0) {
				// no more count until the next enable trigger
				enabled= false;
				director.schedule(doneCommand, null, triggerDelay);
			} // if

			return newValue;
		} // value
	} // Ack
	
	/** set the command to be scheduled when the count is over */
	void setDoneCommand(Command c) {
		doneCommand= c;
	}
	
	/** accessor to startCommand */
	Command getStartCommand() {
		return getEnableTriggerCommand();
	}
	

	/** self check - are all the terminals attached, run the self check from superclass */
	public boolean selfCheck() {
		boolean OK= true;
		boolean check;
		check= (nextTerm != null);
		OK= OK && check;
		if (!check) {
			director.error("Action " + getAlias() + ", id= " + getID()
					+ ", does not have attached terminal "
					+ StrideDefaults.NEXT);
		}

		check= (ackNextTerm != null);
		OK= OK && check;		
		if (!check) {
			director.error("Action " + getAlias() + ", id= " + getID()
					+ ", does not have attached terminal "
					+ StrideDefaults.ACK_NEXT);
		}
		
		check= (counterTerm != null);
		OK= OK && check;		
		if (!check) {
			director.error("Action " + getAlias() + ", id= " + getID()
					+ ", does not have attached terminal "
					+ StrideDefaults.COUNTER);
		}

		check= (ackCounterTerm != null);
		OK= OK && check;		
		if (!check) {
			director.error("Action " + getAlias() + ", id= " + getID()
					+ ", does not have attached terminal "
					+ StrideDefaults.ACK_COUNTER);
		}
		
		// also run the self check on the superclass
		check= super.selfCheck();
		OK= OK && check;
		
		return OK;
	} // selfCheck
	
	
	
} // class CountingAction
